@@ -2,8 +2,13 @@ import functools
 
 import pytest
 import torch
+from utils_fp4 import cast_from_fp4, recover_swizzled_scales, ref_nvfp4_quant
 
-from flashinfer import fp4_quantize, fp4_swizzle_blockscale
+from flashinfer import (
+    e2m1_and_ufp8sf_scale_to_float,
+    fp4_quantize,
+    nvfp4_block_scale_interleave,
+)
 from flashinfer.utils import is_sm100a_supported
 
 DTYPES = [torch.float16, torch.bfloat16]
@@ -15,33 +20,6 @@ CUDA_DEVICES = ["cuda:0"]
 FLOAT4_E2M1_MAX = 6.0
 FLOAT8_E4M3_MAX = torch.finfo(torch.float8_e4m3fn).max
 
-# E2M1 to float
-# 0111 -> 6
-# 0110 -> 4
-# 0101 -> 3
-# 0100 -> 2
-# 0011 -> 1.5
-# 0010 -> 1
-# 0001 -> 0.5
-# 0000 -> 0
-E2M1_TO_FLOAT32 = [
-    0.0,
-    0.5,
-    1.0,
-    1.5,
-    2.0,
-    3.0,
-    4.0,
-    6.0,
-    0.0,
-    -0.5,
-    -1.0,
-    -1.5,
-    -2.0,
-    -3.0,
-    -4.0,
-    -6.0,
-]
 BLOCK_SIZE = 16
 
 
@@ -106,66 +84,6 @@ def unswizzle_sf(
     return sf_unswizzle_sliced.contiguous()
 
 
-def cast_from_fp4(x, m, n):
-    # The fp4 values are packed in uint8 as [v_1st | v_2nd]
-    v_2nd = x & 0xF
-    v_1st = (x >> 4) & 0xF
-    c = torch.stack((v_2nd, v_1st), dim=-1)
-    out = torch.tensor([E2M1_TO_FLOAT32[x] for x in c.flatten()])
-    out = out.reshape(m, n).to(torch.float32)
-    return out
-
-
-def cast_to_fp4(x):
-    sign = torch.sign(x)
-    x = torch.abs(x)
-    x[(x >= 0.0) & (x <= 0.25)] = 0.0
-    x[(x > 0.25) & (x < 0.75)] = 0.5
-    x[(x >= 0.75) & (x <= 1.25)] = 1.0
-    x[(x > 1.25) & (x < 1.75)] = 1.5
-    x[(x >= 1.75) & (x <= 2.5)] = 2.0
-    x[(x > 2.5) & (x < 3.5)] = 3.0
-    x[(x >= 3.5) & (x <= 5.0)] = 4.0
-    x[x > 5.0] = 6.0
-    return x * sign
-
-
-def get_reciprocal(x):
-    if isinstance(x, torch.Tensor):
-        return torch.where(x == 0, torch.tensor(0.0, dtype=x.dtype), 1.0 / x)
-    elif isinstance(x, (float, int)):
-        return 0.0 if x == 0 else 1.0 / x
-    else:
-        raise TypeError("Input must be a float, int, or a torch.Tensor.")
-
-
-def ref_nvfp4_quant(x, global_scale):
-    assert global_scale.dtype == torch.float32
-    assert x.ndim == 2
-    m, n = x.shape
-    x = torch.reshape(x, (m, n // BLOCK_SIZE, BLOCK_SIZE))
-    vec_max = torch.max(torch.abs(x), dim=-1, keepdim=True)[0].to(torch.float32)
-    scale = global_scale * (vec_max * get_reciprocal(FLOAT4_E2M1_MAX))
-    scale = scale.to(torch.float8_e4m3fn).to(torch.float32)
-    output_scale = get_reciprocal(scale * get_reciprocal(global_scale))
-
-    scaled_x = x.to(torch.float32) * output_scale
-    clipped_x = torch.clamp(scaled_x, -6.0, 6.0).reshape(m, n)
-    return cast_to_fp4(clipped_x), scale.squeeze(-1)
-
-
-def recover_swizzled_scales(scale, m, n):
-    round_up = lambda x, y: (x + y - 1) // y * y
-    rounded_m = round_up(m, 128)
-    scale_n = n // BLOCK_SIZE
-    rounded_n = round_up(scale_n, 4)
-    # Recover the swizzled scaling factor to linear layout
-    tmp = torch.reshape(scale, (1, rounded_m // 128, rounded_n // 4, 32, 4, 4))
-    tmp = torch.permute(tmp, (0, 1, 4, 3, 2, 5))
-    result = torch.reshape(tmp, (rounded_m, rounded_n)).to(torch.float32)
-    return result[:m, :scale_n]
-
-
 @pytest.mark.parametrize("dtype", DTYPES)
 @pytest.mark.parametrize("shape", SHAPES)
 @pytest.mark.parametrize("seed", SEEDS)
@@ -185,14 +103,16 @@ def test_fp4_quantization(
     x = torch.randn((m, n), dtype=dtype)
     tensor_amax = torch.abs(x).max().to(torch.float32)
     global_scale = FLOAT8_E4M3_MAX * FLOAT4_E2M1_MAX / tensor_amax
-    out_ref, scale_ref = ref_nvfp4_quant(x, global_scale)
-
+    out_ref, scale_ref = ref_nvfp4_quant(x, global_scale, BLOCK_SIZE)
     out, out_scale = fp4_quantize(x, global_scale, BLOCK_SIZE, False)
     assert n % BLOCK_SIZE == 0, f"cols needs to be {BLOCK_SIZE} divisible"
     scale_ans = recover_swizzled_scales(
-        out_scale.reshape(-1, n // BLOCK_SIZE).view(torch.float8_e4m3fn), m, n
+        out_scale.reshape(-1, n // BLOCK_SIZE).view(torch.float8_e4m3fn),
+        m,
+        n,
+        BLOCK_SIZE,
     )
-    out_ans = cast_from_fp4(out, m, n)
+    out_ans = cast_from_fp4(out).reshape(m, n)
     torch.testing.assert_close(out_ans, out_ref, rtol=1e0, atol=1e-1)
     torch.testing.assert_close(scale_ans, scale_ref, rtol=1e-1, atol=1e-1)
 
@@ -238,12 +158,12 @@ def test_scale_swizzling(
 @pytest.mark.parametrize("seed", SEEDS)
 @pytest.mark.parametrize("device", CUDA_DEVICES)
 @torch.inference_mode()
-def test_fp4_swizzle_blockscale(
+def test_nvfp4_block_scale_interleave(
     shape: tuple[int, int],
     seed: int,
     device: str,
 ) -> None:
-    """Test the fp4_swizzle_blockscale function directly."""
+    """Test the nvfp4_block_scale_interleave function directly."""
     if not is_sm100a_supported(torch.device("cuda")):
         pytest.skip("Nvfp4 Requires compute capability of 10 or above")
     torch.set_default_device(device)
@@ -258,7 +178,7 @@ def test_fp4_swizzle_blockscale(
     unswizzled_sf = torch.randint(0, 256, scale_shape, dtype=torch.uint8, device=device)
 
     # Test the swizzling function
-    swizzled_sf = fp4_swizzle_blockscale(unswizzled_sf, m, n, sf_vec_size)
+    swizzled_sf = nvfp4_block_scale_interleave(unswizzled_sf)
 
     # Compare against the reference implementation
     ref_swizzled_sf = swizzle_sf(unswizzled_sf, m, n, sf_vec_size)
@@ -272,12 +192,88 @@ def test_fp4_swizzle_blockscale(
     padded_row = ((m + 128 - 1) // 128) * 128  # Next multiple of 128
     padded_col = ((n + factor - 1) // factor) * factor  # Next multiple of 64
     expected_shape = (padded_row, padded_col // sf_vec_size)
+    expected_size = expected_shape[0] * expected_shape[1]
 
     assert (
-        swizzled_sf.shape == expected_shape
-    ), f"Expected shape {expected_shape}, got {swizzled_sf.shape}"
+        expected_size == swizzled_sf.shape[0]
+    ), f"Expected size {expected_size}, got {swizzled_sf.shape[0]}"
     assert_equal = functools.partial(torch.testing.assert_close, rtol=0, atol=0)
-    assert_equal(swizzled_sf, ref_swizzled_sf)
+    assert_equal(swizzled_sf.reshape(expected_shape), ref_swizzled_sf)
+
+
+@pytest.mark.parametrize("shape", SHAPES)
+@pytest.mark.parametrize("seed", SEEDS)
+@pytest.mark.parametrize("device", CUDA_DEVICES)
+@pytest.mark.parametrize("sf_use_ue8m0", [True, False])
+@torch.inference_mode()
+def test_e2m1_dequantization(
+    shape: tuple[int, int],
+    seed: int,
+    device: str,
+    sf_use_ue8m0: bool,
+) -> None:
+    """Test roundtrip: fp4_quantize -> e2m1_and_ufp8sf_scale_to_float."""
+    if not is_sm100a_supported(torch.device("cuda")):
+        pytest.skip("Nvfp4 Requires compute capability of 10 or above")
+    torch.set_default_device(device)
+    torch.manual_seed(seed)
+
+    # Create a reasonable test tensor
+    m, n = shape
+    x = torch.randn((m, n), dtype=torch.float16)
+
+    # Calculate global scale as in the other tests
+    tensor_amax = torch.abs(x).max().to(torch.float32)
+    global_scale = FLOAT8_E4M3_MAX * FLOAT4_E2M1_MAX / tensor_amax
+
+    # Test with default common settings
+    is_sf_swizzled_layout = True
+    block_size = 32 if sf_use_ue8m0 else 16
+
+    # Step 1: Quantize with fp4_quantize
+    quantized_tensor, scale_factors = fp4_quantize(
+        x, global_scale, block_size, sf_use_ue8m0, is_sf_swizzled_layout
+    )
+
+    # Step 2: Dequantize with e2m1_and_ufp8sf_scale_to_float
+    ufp8_type = 0 if sf_use_ue8m0 else 1
+    dequantized_tensor = e2m1_and_ufp8sf_scale_to_float(
+        quantized_tensor,
+        scale_factors,
+        1 / global_scale,
+        sf_vec_size=block_size,
+        ufp8_type=ufp8_type,
+        is_sf_swizzled_layout=is_sf_swizzled_layout,
+    )
+
+    # Move back to device for comparison
+    dequantized_tensor = dequantized_tensor.to(device)
+    x_float32 = x.to(torch.float32)
+
+    # Step 3: Compare results
+    assert (
+        dequantized_tensor.shape == x.shape
+    ), f"Shape mismatch: expected {x.shape}, got {dequantized_tensor.shape}"
+    assert (
+        dequantized_tensor.dtype == torch.float32
+    ), f"Expected float32, got {dequantized_tensor.dtype}"
+
+    # Check for invalid values
+    assert not torch.isnan(
+        dequantized_tensor
+    ).any(), "Dequantized tensor contains NaN values"
+    assert not torch.isinf(
+        dequantized_tensor
+    ).any(), "Dequantized tensor contains Inf values"
+
+    # Compare with original - should be reasonably close since FP4 is designed to preserve important values
+    torch.testing.assert_close(
+        dequantized_tensor,
+        x_float32,
+        rtol=0.3,
+        atol=0.5,  # Reasonable tolerance for FP4 quantization
+        msg="Quantize -> dequantize roundtrip failed",
+    )
 
 
 if __name__ == "__main__":
